package network

import (
	"encoding/json"
	"errors"
	"fmt"
	"regexp"
	"time"

	"BFT/pbft/consensus"
)

type Node struct {
	NodeID        string
	NodeTable     map[string]string // key=nodeID, value=url
	View          *View
	CurrentState  *consensus.State
	CommittedMsgs []*consensus.RequestMsg // kinda block.
	MsgBuffer     *MsgBuffer
	MsgEntrance   chan interface{}
	MsgDelivery   chan interface{}
	Alarm         chan bool
}

type MsgBuffer struct {
	ReqMsgs           []*consensus.RequestMsg
	PrePrepareMsgs    []*consensus.PrePrepareMsg
	PrepareMsgs       []*consensus.VoteMsg
	CommitMsgs        []*consensus.VoteMsg
	GlobalForwardMsgs []*consensus.GlobalForwardMsg
	LocalForwardMsgs  []*consensus.LocalForwardMsg
}

type View struct {
	ID      int64
	Primary string
}

const ResolvingTimeDuration = time.Millisecond * 1000 // 1 second.
const f = 1                                           // only 1 byzantine node in each shard

func NewNode(nodeID string) *Node {
	const viewID = 10000000000 // temporary.

	node := &Node{
		// Hard-coded for test.
		NodeID: nodeID,
		NodeTable: map[string]string{
			"p1_0": "localhost:1111",
			"r1_1": "localhost:1112",
			"r1_2": "localhost:1113",
			"r1_3": "localhost:1114",
			"p2_0": "localhost:1115",
			"r2_1": "localhost:1116",
			"r2_2": "localhost:1117",
			"r2_3": "localhost:1118",
		},
		View: &View{
			ID:      viewID,
			Primary: "p1",
		},

		// Consensus-related struct
		CurrentState:  nil,
		CommittedMsgs: make([]*consensus.RequestMsg, 0),
		MsgBuffer: &MsgBuffer{
			ReqMsgs:           make([]*consensus.RequestMsg, 0),
			PrePrepareMsgs:    make([]*consensus.PrePrepareMsg, 0),
			PrepareMsgs:       make([]*consensus.VoteMsg, 0),
			CommitMsgs:        make([]*consensus.VoteMsg, 0),
			GlobalForwardMsgs: make([]*consensus.GlobalForwardMsg, 0),
		},

		// Channels
		MsgEntrance: make(chan interface{}),
		MsgDelivery: make(chan interface{}),
		Alarm:       make(chan bool),
	}

	// Start message dispatcher
	go node.dispatchMsg()

	// Start alarm trigger
	go node.alarmToDispatcher()

	// Start message resolver
	go node.resolveMsg()

	return node
}

func (node *Node) Broadcast(msg interface{}, path string) map[string]error {
	errorMap := make(map[string]error)

	for nodeID, url := range node.NodeTable {
		if nodeID == node.NodeID {
			continue
		}

		// Find the shard number through regex
		// Only broadcast to other replicas in the same shard
		re := regexp.MustCompile(`[0-9]+`)
		shardNumTo, shardNumFrom := int(re.Find([]byte(nodeID))[0])-48, int(re.Find([]byte(node.NodeID))[0])-48
		if shardNumFrom != shardNumTo {
			continue
		}
		// fmt.Printf("ShardNumFrom is %d, ShardNumTo is %d", shardNumFrom, shardNumTo)

		jsonMsg, err := json.Marshal(msg)
		if err != nil {
			errorMap[nodeID] = err
			continue
		}

		send(url+path, jsonMsg)
	}

	if len(errorMap) == 0 {
		return nil
	} else {
		return errorMap
	}
}

func (node *Node) GlobalShare(msg *consensus.GlobalForwardMsg, path string) map[string]error {
	errorMap := make(map[string]error)

	// Get the shard num of the source
	re := regexp.MustCompile(`[0-9]+`)
	shardNumFrom, shardNumTo := int(re.Find([]byte(node.NodeID))[0])-48, 0

	// Get the shard num of the target
	for index, ringOrder := range msg.ReqMsg.RingOrder {
		if ringOrder == shardNumFrom {
			if index != len(msg.ReqMsg.RingOrder)-1 {
				shardNumTo = msg.ReqMsg.RingOrder[(index+1)%len(msg.ReqMsg.RingOrder)]
			} else {
				fmt.Printf("[Lock Finish] This is the last shard in Global sharing\n")
				fmt.Printf("[Execute Transactions]\n")
				node.CurrentState.CurrentStage = consensus.Executed
				shardNumTo = msg.ReqMsg.RingOrder[(index+1)%len(msg.ReqMsg.RingOrder)]
			}
		}
	}

	for nodeID, url := range node.NodeTable {
		// Not in the next shard
		if int(re.Find([]byte(nodeID))[0])-48 != shardNumTo {
			continue
		}
		// Not with the same id
		if int(re.FindAll([]byte(nodeID), 2)[1][0]) != int(re.FindAll([]byte(node.NodeID), 2)[1][0]) {
			continue
		}
		// Encode GlobalForwardMsg
		jsonMsg, err := json.Marshal(msg)
		if err != nil {
			errorMap[nodeID] = err
			fmt.Printf("Error\n")
			continue
		}

		send(url+path, jsonMsg)
		break
	}

	if len(errorMap) == 0 {
		return nil
	} else {
		return errorMap
	}
}

func (node *Node) Reply(msg *consensus.ReplyMsg) error {
	// Print all committed messages.
	for _, value := range node.CommittedMsgs {
		fmt.Printf("Committed value: %s, %d, %s, %d\n", value.ClientID, value.Timestamp, value.Operation, value.SequenceID)
	}

	jsonMsg, err := json.Marshal(msg)
	if err != nil {
		return err
	}

	send(node.NodeTable[node.View.Primary]+"/reply", jsonMsg)

	return nil
}

// GetReq can be called when the node's CurrentState is nil.
// Consensus start procedure for the Primary.
func (node *Node) GetReq(reqMsg *consensus.RequestMsg) error {
	LogMsg(reqMsg)

	// Create a new state for the new consensus.
	err := node.createStateForNewConsensus()
	if err != nil {
		return err
	}

	// Start the consensus process.
	prePrepareMsg, err := node.CurrentState.StartConsensus(reqMsg)
	if err != nil {
		return err
	}

	LogStage(fmt.Sprintf("Consensus Process (ViewID:%d)", node.CurrentState.ViewID), false)

	// Send getPrePrepare message
	if prePrepareMsg != nil {
		node.Broadcast(prePrepareMsg, "/preprepare")
		LogStage("Pre-prepare", true)
	}

	return nil
}

// GetPrePrepare can be called when the node's CurrentState is nil.
// Consensus start procedure for normal participants.
func (node *Node) GetPrePrepare(prePrepareMsg *consensus.PrePrepareMsg) error {
	LogMsg(prePrepareMsg)

	// Create a new state for the new consensus.
	err := node.createStateForNewConsensus()
	if err != nil {
		return err
	}

	prePareMsg, err := node.CurrentState.PrePrepare(prePrepareMsg)
	if err != nil {
		return err
	}

	if prePareMsg != nil {
		// Attach node ID to the message
		prePareMsg.NodeID = node.NodeID

		LogStage("Pre-prepare", true)
		node.Broadcast(prePareMsg, "/prepare")
		LogStage("Prepare", false)
	}

	return nil
}

func (node *Node) GetPrepare(prepareMsg *consensus.VoteMsg) error {
	LogMsg(prepareMsg)

	commitMsg, err := node.CurrentState.Prepare(prepareMsg)
	if err != nil {
		return err
	}

	if commitMsg != nil {
		// Attach node ID to the message
		commitMsg.NodeID = node.NodeID

		LogStage("Prepare", true)
		node.Broadcast(commitMsg, "/commit")
		LogStage("Commit", false)
	}

	return nil
}

func (node *Node) GetCommit(commitMsg *consensus.VoteMsg) error {
	LogMsg(commitMsg)

	// Check the state
	replyMsg, committedMsg, globalForwardMsg, err := node.CurrentState.Commit(commitMsg)
	if err != nil {
		return err
	}

	if replyMsg != nil {
		if committedMsg == nil {
			return errors.New("committed message is nil, even though the reply message is not nil")
		}

		// Attach node ID to the message
		replyMsg.NodeID = node.NodeID

		// Save the last version of committed messages to node.
		node.CommittedMsgs = append(node.CommittedMsgs, committedMsg)

		LogStage("Commit", true)
		node.Reply(replyMsg)
		LogStage("Reply", true)
	}

	if globalForwardMsg != nil {
		LogStage("Global Sharing", false)
		node.GlobalShare(globalForwardMsg, "/globalforward")
		LogStage("Global Sharing", true)
	}

	return nil
}

func (node *Node) GetReply(msg *consensus.ReplyMsg) {
	fmt.Printf("Result: %s by %s\n", msg.Result, msg.NodeID)
}

// TODO: Verify msg
// FIXME: Ugly code
func (node *Node) GetGlobalForward(msg *consensus.GlobalForwardMsg) error {
	// Transactions have already been executed
	if node.CurrentState != nil && node.CurrentState.CurrentStage == consensus.Executed {
		return nil
	}

	LogMsg(msg)

	if node.NodeID[0] == "p"[0] {
		// Primary
		if node.CurrentState == nil {
			err := node.createStateForNewConsensus()
			if err != nil {
				return err
			}
			node.CurrentState.MsgLogs.GlobalForwardMsgs[msg.CommitMsg.NodeID] = msg
		} else if node.CurrentState.CurrentStage == consensus.Committed {
			fmt.Printf("Execute Transaction\n")
			node.GlobalShare(msg, "/globalforward")
			return nil
		} else {
			node.CurrentState.MsgLogs.GlobalForwardMsgs[msg.CommitMsg.NodeID] = msg
			if len(node.CurrentState.MsgLogs.GlobalForwardMsgs) >= 2*f {
				node.CurrentState.CurrentStage = consensus.Idle
				prePrepareMsg, err := node.CurrentState.StartConsensus(msg.ReqMsg)
				if err != nil {
					return err
				}

				node.CurrentState.MsgLogs.ReqMsg = msg.ReqMsg
				node.Broadcast(prePrepareMsg, "/preprepare")
				LogStage("Pre-prepare", true)
				return nil
			}
		}
	} else {
		// Replica
		if node.CurrentState != nil && node.CurrentState.CurrentStage == consensus.Committed {
			fmt.Printf("Execute Transaction\n")
			node.GlobalShare(msg, "/globalforward")
			return nil
		}
		LogStage("Local Sharing", false)
		localForwardMsg := &consensus.LocalForwardMsg{
			GlobalForwardMsg: msg,
		}
		node.Broadcast(localForwardMsg, "/localforward")
		LogStage("Local Sharing", true)
	}
	return nil
}

func (node *Node) GetLocalForward(msg *consensus.LocalForwardMsg) error {
	LogMsg(msg)
	if node.NodeID[0] == "p"[0] {
		// Primary
		if node.CurrentState == nil {
			err := node.createStateForNewConsensus()
			if err != nil {
				return err
			}
			node.CurrentState.MsgLogs.GlobalForwardMsgs[msg.GlobalForwardMsg.CommitMsg.NodeID] = msg.GlobalForwardMsg
		} else {
			node.CurrentState.MsgLogs.GlobalForwardMsgs[msg.GlobalForwardMsg.CommitMsg.NodeID] = msg.GlobalForwardMsg
			if len(node.CurrentState.MsgLogs.GlobalForwardMsgs) >= 2*f {
				node.CurrentState.CurrentStage = consensus.Idle
				prePrepareMsg, err := node.CurrentState.StartConsensus(msg.GlobalForwardMsg.ReqMsg)
				if err != nil {
					return err
				}

				node.CurrentState.MsgLogs.ReqMsg = msg.GlobalForwardMsg.ReqMsg
				node.Broadcast(prePrepareMsg, "/preprepare")
				LogStage("Pre-prepare", true)
				return nil
			}
		}
	}
	return nil
}

func (node *Node) createStateForNewConsensus() error {
	// Check if there is an ongoing consensus process.
	if node.CurrentState != nil {
		return errors.New("another consensus is ongoing")
	}

	// Get the last sequence ID
	var lastSequenceID int64
	if len(node.CommittedMsgs) == 0 {
		lastSequenceID = -1
	} else {
		lastSequenceID = node.CommittedMsgs[len(node.CommittedMsgs)-1].SequenceID
	}

	// Create a new state for this new consensus process in the Primary
	node.CurrentState = consensus.CreateState(node.View.ID, lastSequenceID)

	LogStage("Create the replica status", true)

	return nil
}

func (node *Node) dispatchMsg() {
	for {
		select {
		case msg := <-node.MsgEntrance:
			err := node.routeMsg(msg)
			if err != nil {
				fmt.Println(err)
				// TODO: send err to ErrorChannel
			}
		case <-node.Alarm:
			err := node.routeMsgWhenAlarmed()
			if err != nil {
				fmt.Println(err)
				// TODO: send err to ErrorChannel
			}
		}
	}
}

func (node *Node) routeMsg(msg interface{}) []error {
	switch msg.(type) {
	case *consensus.RequestMsg:
		if node.CurrentState == nil {
			// Copy buffered messages first.
			msgs := make([]*consensus.RequestMsg, len(node.MsgBuffer.ReqMsgs))
			copy(msgs, node.MsgBuffer.ReqMsgs)

			// Append a newly arrived message.
			msgs = append(msgs, msg.(*consensus.RequestMsg))

			// Empty the buffer.
			node.MsgBuffer.ReqMsgs = make([]*consensus.RequestMsg, 0)

			// Send messages.
			node.MsgDelivery <- msgs
		} else {
			node.MsgBuffer.ReqMsgs = append(node.MsgBuffer.ReqMsgs, msg.(*consensus.RequestMsg))
		}
	case *consensus.PrePrepareMsg:
		if node.CurrentState == nil {
			// Copy buffered messages first.
			msgs := make([]*consensus.PrePrepareMsg, len(node.MsgBuffer.PrePrepareMsgs))
			copy(msgs, node.MsgBuffer.PrePrepareMsgs)

			// Append a newly arrived message.
			msgs = append(msgs, msg.(*consensus.PrePrepareMsg))

			// Empty the buffer.
			node.MsgBuffer.PrePrepareMsgs = make([]*consensus.PrePrepareMsg, 0)

			// Send messages.
			node.MsgDelivery <- msgs
		} else {
			node.MsgBuffer.PrePrepareMsgs = append(node.MsgBuffer.PrePrepareMsgs, msg.(*consensus.PrePrepareMsg))
		}
	case *consensus.VoteMsg:
		if msg.(*consensus.VoteMsg).MsgType == consensus.PrepareMsg {
			if node.CurrentState == nil || node.CurrentState.CurrentStage != consensus.PrePrepared {
				node.MsgBuffer.PrepareMsgs = append(node.MsgBuffer.PrepareMsgs, msg.(*consensus.VoteMsg))
			} else {
				// Copy buffered messages first.
				msgs := make([]*consensus.VoteMsg, len(node.MsgBuffer.PrepareMsgs))
				copy(msgs, node.MsgBuffer.PrepareMsgs)

				// Append a newly arrived message.
				msgs = append(msgs, msg.(*consensus.VoteMsg))

				// Empty the buffer.
				node.MsgBuffer.PrepareMsgs = make([]*consensus.VoteMsg, 0)

				// Send messages.
				node.MsgDelivery <- msgs
			}
		} else if msg.(*consensus.VoteMsg).MsgType == consensus.CommitMsg {
			if node.CurrentState == nil || node.CurrentState.CurrentStage != consensus.Prepared {
				node.MsgBuffer.CommitMsgs = append(node.MsgBuffer.CommitMsgs, msg.(*consensus.VoteMsg))
			} else {
				// Copy buffered messages first.
				msgs := make([]*consensus.VoteMsg, len(node.MsgBuffer.CommitMsgs))
				copy(msgs, node.MsgBuffer.CommitMsgs)

				// Append a newly arrived message.
				msgs = append(msgs, msg.(*consensus.VoteMsg))

				// Empty the buffer.
				node.MsgBuffer.CommitMsgs = make([]*consensus.VoteMsg, 0)

				// Send messages.
				node.MsgDelivery <- msgs
			}
		}
	case *consensus.GlobalForwardMsg:
		// TODO: Handle different current states
		if node.CurrentState == nil {
			msgs := make([]*consensus.GlobalForwardMsg, len(node.MsgBuffer.GlobalForwardMsgs))
			copy(msgs, node.MsgBuffer.GlobalForwardMsgs)
			msgs = append(msgs, msg.(*consensus.GlobalForwardMsg))
			node.MsgBuffer.GlobalForwardMsgs = make([]*consensus.GlobalForwardMsg, 0)

			node.MsgDelivery <- msgs
		}
	case *consensus.LocalForwardMsg:
		msgs := make([]*consensus.LocalForwardMsg, len(node.MsgBuffer.LocalForwardMsgs))
		copy(msgs, node.MsgBuffer.LocalForwardMsgs)
		msgs = append(msgs, msg.(*consensus.LocalForwardMsg))
		node.MsgBuffer.LocalForwardMsgs = make([]*consensus.LocalForwardMsg, 0)

		node.MsgDelivery <- msgs
	}
	return nil
}

func (node *Node) routeMsgWhenAlarmed() []error {
	if node.CurrentState == nil {
		// Check ReqMsgs, send them.
		if len(node.MsgBuffer.ReqMsgs) != 0 {
			msgs := make([]*consensus.RequestMsg, len(node.MsgBuffer.ReqMsgs))
			copy(msgs, node.MsgBuffer.ReqMsgs)

			node.MsgDelivery <- msgs
		}

		// Check PrePrepareMsgs, send them.
		if len(node.MsgBuffer.PrePrepareMsgs) != 0 {
			msgs := make([]*consensus.PrePrepareMsg, len(node.MsgBuffer.PrePrepareMsgs))
			copy(msgs, node.MsgBuffer.PrePrepareMsgs)

			node.MsgDelivery <- msgs
		}
	} else {
		switch node.CurrentState.CurrentStage {
		case consensus.PrePrepared:
			// Check PrepareMsgs, send them.
			if len(node.MsgBuffer.PrepareMsgs) != 0 {
				msgs := make([]*consensus.VoteMsg, len(node.MsgBuffer.PrepareMsgs))
				copy(msgs, node.MsgBuffer.PrepareMsgs)

				node.MsgDelivery <- msgs
			}
		case consensus.Prepared:
			// Check CommitMsgs, send them.
			if len(node.MsgBuffer.CommitMsgs) != 0 {
				msgs := make([]*consensus.VoteMsg, len(node.MsgBuffer.CommitMsgs))
				copy(msgs, node.MsgBuffer.CommitMsgs)

				node.MsgDelivery <- msgs
			}
		}
	}

	return nil
}

func (node *Node) resolveMsg() {
	for {
		// Get buffered messages from the dispatcher.
		msgs := <-node.MsgDelivery
		switch msgs.(type) {
		case []*consensus.RequestMsg:
			errs := node.resolveRequestMsg(msgs.([]*consensus.RequestMsg))
			if len(errs) != 0 {
				for _, err := range errs {
					fmt.Println(err)
				}
				// TODO: send err to ErrorChannel
			}
		case []*consensus.PrePrepareMsg:
			errs := node.resolvePrePrepareMsg(msgs.([]*consensus.PrePrepareMsg))
			if len(errs) != 0 {
				for _, err := range errs {
					fmt.Println(err)
				}
				// TODO: send err to ErrorChannel
			}
		case []*consensus.VoteMsg:
			voteMsgs := msgs.([]*consensus.VoteMsg)
			if len(voteMsgs) == 0 {
				break
			}

			if voteMsgs[0].MsgType == consensus.PrepareMsg {
				errs := node.resolvePrepareMsg(voteMsgs)
				if len(errs) != 0 {
					for _, err := range errs {
						fmt.Println(err)
					}
					// TODO: send err to ErrorChannel
				}
			} else if voteMsgs[0].MsgType == consensus.CommitMsg {
				errs := node.resolveCommitMsg(voteMsgs)
				if len(errs) != 0 {
					for _, err := range errs {
						fmt.Println(err)
					}
					// TODO: send err to ErrorChannel
				}
			}
		case []*consensus.GlobalForwardMsg:
			errs := node.resolveGlobalForwardMsg(msgs.([]*consensus.GlobalForwardMsg))
			if len(errs) != 0 {
				for _, err := range errs {
					fmt.Println(err)
				}
			}
		case []*consensus.LocalForwardMsg:
			errs := node.resolveLocalForwardMsg(msgs.([]*consensus.LocalForwardMsg))
			if len(errs) != 0 {
				for _, err := range errs {
					fmt.Println(err)
				}
			}
		}
	}
}

func (node *Node) alarmToDispatcher() {
	for {
		time.Sleep(ResolvingTimeDuration)
		node.Alarm <- true
	}
}

func (node *Node) resolveRequestMsg(msgs []*consensus.RequestMsg) []error {
	errs := make([]error, 0)

	// Resolve messages
	for _, reqMsg := range msgs {
		err := node.GetReq(reqMsg)
		if err != nil {
			errs = append(errs, err)
		}
	}

	if len(errs) != 0 {
		return errs
	}

	return nil
}

func (node *Node) resolvePrePrepareMsg(msgs []*consensus.PrePrepareMsg) []error {
	errs := make([]error, 0)

	// Resolve messages
	for _, prePrepareMsg := range msgs {
		err := node.GetPrePrepare(prePrepareMsg)
		if err != nil {
			errs = append(errs, err)
		}
	}

	if len(errs) != 0 {
		return errs
	}

	return nil
}

func (node *Node) resolvePrepareMsg(msgs []*consensus.VoteMsg) []error {
	errs := make([]error, 0)

	// Resolve messages
	for _, prepareMsg := range msgs {
		err := node.GetPrepare(prepareMsg)
		if err != nil {
			errs = append(errs, err)
		}
	}

	if len(errs) != 0 {
		return errs
	}

	return nil
}

func (node *Node) resolveCommitMsg(msgs []*consensus.VoteMsg) []error {
	errs := make([]error, 0)

	// Resolve messages
	for _, commitMsg := range msgs {
		err := node.GetCommit(commitMsg)
		if err != nil {
			errs = append(errs, err)
		}
	}

	if len(errs) != 0 {
		return errs
	}

	return nil
}

func (node *Node) resolveGlobalForwardMsg(msgs []*consensus.GlobalForwardMsg) []error {
	errs := make([]error, 0)

	for _, globalForwardMsg := range msgs {
		err := node.GetGlobalForward(globalForwardMsg)
		if err != nil {
			errs = append(errs, err)
		}
	}

	if len(errs) != 0 {
		return errs
	}

	return nil
}

func (node *Node) resolveLocalForwardMsg(msgs []*consensus.LocalForwardMsg) []error {
	errs := make([]error, 0)

	for _, localForwardMsg := range msgs {
		err := node.GetLocalForward(localForwardMsg)
		if err != nil {
			errs = append(errs, err)
		}
	}

	if len(errs) != 0 {
		return errs
	}

	return nil
}
